@@ -9,21 +9,23 @@ from ipywidgets import Play, Checkbox, IntSlider, Button, ColorPicker, VBox, Tab
 from k3d.plot import Plot
 
 from .jointset import JointSet, MuPoTSJoints, OpenPoseJoints, CocoExJoints, PanopticJoints, Common14Joints
+from .loss import L2
 from .objects import Skeleton, DrawableSkeleton, Color, Positions
 
 
 class VideoPlayer:
-    def __init__(self, plot: Plot):
+    def __init__(self, plot: Plot, fps: int, frames: np.ndarray, video_player: Play = None):
         self.plot: Plot = plot
-        self.video_player: Optional[Play] = None
-
-    def get_video_player_widget(self, fps: int, frames: np.ndarray) -> HBox:
-        self.video_player = Play(
+        self.video_player: Optional[Play] = video_player if video_player is not None else Play(
             value=0,
             min=0, max=frames.shape[0] - 1,
             step=1, interval=1000 / fps,
             description='Press play', disabled=False)
-        frame_slider = IntSlider(value=0, min=0, max=frames.shape[0] - 1, step=1, description='Frame')
+        self.fps: int = fps
+        self.frames: np.ndarray = frames
+
+    def get_video_player_widget(self) -> HBox:
+        frame_slider = IntSlider(value=0, min=0, max=self.frames.shape[0] - 1, step=1, description='Frame')
         next_frame_button = Button(description='Next frame')
         next_frame_button.on_click(self.__update_to_next_frame)
         previous_frame_button = Button(description='Previous frame')
@@ -182,6 +184,18 @@ class ColorChanger:
         return '{0:#0{1}x}'.format(color, 8).replace('0x', '#')
 
 
+class LossContainer:
+    def __init__(self, pred_skeletons: np.ndarray, gt_skeletons: np.ndarray,
+                 loss: Callable[[np.ndarray, np.ndarray], float]):
+        self.pred_skeletons: np.ndarray = pred_skeletons
+        self.gt_skeletons: np.ndarray = gt_skeletons
+        self.loss: Callable[[np.ndarray, np.ndarray], float] = loss
+
+    def get_loss_tab(self) -> Tab:
+        self.loss = lambda x, y: x + y
+        return Tab()
+
+
 class SkeletonVisualizer:
     """ A 3D skeleton visualizer which can visualize skeletons in 3D plots."""
 
@@ -196,7 +210,7 @@ class SkeletonVisualizer:
     def visualize(
             self, skeletons: np.ndarray, colors: List[Color] = None, include_names: bool = False,
             include_coordinates: bool = False, automatic_camera_orientation: bool = False,
-            is_gt_list: List[bool] = None) -> None:
+            is_gt_list: List[bool] = None, additional_tabs: List[Tuple[str, Widget]] = None) -> None:
         self.__assert_include_arguments(include_names, include_coordinates)
         self.__assert_skeleton_shapes(skeletons)
         colors = self.__init_colors(skeletons.shape[0], colors)
@@ -214,13 +228,14 @@ class SkeletonVisualizer:
         color_changer: ColorChanger = ColorChanger()
         color_changer_tab: Tab = color_changer.get_color_changer_widget(self.skeletons)
         self.__display_interface([('Change visibilities', visibility_widget),
-                                  ('Change colors', color_changer_tab)])
+                                  ('Change colors', color_changer_tab)]
+                                 + [] if additional_tabs is None else additional_tabs)
 
     def visualize_with_ground_truths(
             self, pred_skeletons: np.ndarray, gt_skeletons: np.ndarray,
             pred_colors: List[Color] = None, gt_colors: List[Color] = None,
             include_names: bool = False, include_coordinates: bool = False,
-            automatic_camera_orientation: bool = False) -> None:
+            automatic_camera_orientation: bool = False, loss: Callable[[np.ndarray, np.ndarray], float] = L2) -> None:
         assert pred_skeletons.shape == gt_skeletons.shape, \
             'The predicate and ground truth skeleton arrays must have the same shape.'
         pred_colors = self.__init_colors(pred_skeletons.shape[0], pred_colors)
@@ -228,13 +243,16 @@ class SkeletonVisualizer:
         skeletons: np.ndarray = np.concatenate((pred_skeletons, gt_skeletons), axis=0)
         colors: List[Color] = pred_colors + gt_colors
         is_gt_list: List[bool] = [False] * len(pred_skeletons) + [True] * len(gt_skeletons)
+        loss_container: LossContainer = LossContainer(pred_skeletons, gt_skeletons, loss)
         self.visualize(
-            skeletons, colors, include_names, include_coordinates, automatic_camera_orientation, is_gt_list)
+            skeletons, colors, include_names, include_coordinates, automatic_camera_orientation,
+            is_gt_list, additional_tabs=[('Losses', loss_container.get_loss_tab())])
 
     def visualize_video(
             self, frames: np.ndarray, colors: List[Color] = None, fps: int = 15,
             include_names: bool = False, include_coordinates: bool = False,
-            automatic_camera_orientation: bool = False, is_gt_list: List[bool] = None) -> None:
+            automatic_camera_orientation: bool = False, is_gt_list: List[bool] = None,
+            additional_tabs: List[Tuple[str, Widget]] = None) -> None:
         self.__assert_include_arguments(include_names, include_coordinates)
         assert len(frames.shape) == 4
         first_frame: np.ndarray = frames[0]
@@ -253,13 +271,14 @@ class SkeletonVisualizer:
         self.__link_text_widgets(include_names, include_coordinates)
         self.plot.display()
         visibility_widget: HBox = HBox([self.joint_names_visible, self.joint_coordinates_visible])
-        video_player: VideoPlayer = VideoPlayer(self.plot)
-        video_player_widget: HBox = video_player.get_video_player_widget(fps, frames)
+        video_player: VideoPlayer = VideoPlayer(self.plot, fps, frames)
+        video_player_widget: HBox = video_player.get_video_player_widget()
         color_changer: ColorChanger = ColorChanger()
         color_changer_tab: Tab = color_changer.get_color_changer_widget(self.skeletons)
-        self.__display_interface([('Change visibilities', visibility_widget),
-                                  ('Change colors', color_changer_tab),
-                                  ('Play video', video_player_widget)])
+        self.__display_interface([('Play video', video_player_widget),
+                                  ('Change visibilities', visibility_widget),
+                                  ('Change colors', color_changer_tab)]
+                                 + [] if additional_tabs is None else additional_tabs)
 
     def visualize_video_from_file(
             self, file_name: str, colors: List[Color] = None, fps: int = 15,
